#include "fastlz.h"
#include "filesystem.h"
#include "nds/arm9/background.h"
#include "nds/arm9/console.h"
#include "nds/cothread.h"
#include "nds/dma.h"
#include "nds/timers.h"
#include "raylibds.hpp"
#include <cstdint>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <nds.h>
#include <maxmod9.h>

#define PALETTE_SIZE 256
#define FRAME_SIZE 256*192

// TODO: hardcode fix
#define CHUNK_SIZE 10
#define QUEUE_SIZE 6
#define FRAMERATE 30
#define VIDEO_TOTALFRAME 827

#define MMSTREAM_BUF_SIZE 800
#define MUSIC_BUFFER_SIZE 16000

void wait_forever(const char* msg);
void LoadNextChunk();
void TimerCallback();
void VBLCallback();
void FrameStep();
int ThreadEntrypoint(void *arg);
int ThreadVideo(void *arg);
mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format );

struct RingBuffer {
	char *buffer = nullptr;
	int head;
	int tail;
	int size;
	int count;

	RingBuffer(size_t size_t)
	{
		buffer = (char*)malloc(size_t);
		size = size_t;
		head = 0;
		tail = 0;
		count = 0;
	}

	~RingBuffer()
	{
		free(buffer);
	}

	int read(char *dest, size_t len)
	{
		if (count >= len)
		{
			count -= len;
			int linear_len = size - head;

			if (len < linear_len)
				memcpy(dest, &buffer[head], len);
				// swiCopy(&buffer[head], dest, len);
			else
			{
				memcpy(dest, &buffer[head], linear_len);
				memcpy(&dest[linear_len], buffer, len-linear_len);
				// swiCopy(&buffer[head], dest, linear_len);
				// swiCopy(buffer, &dest[linear_len], len-linear_len);
			}

			head = (head+len) % size;

			return 1;
		}

		return 0;
	}

	int write(char *data, size_t len)
	{
		if (count+len <= size)
		{
			count += len;
			int linear_len = size - tail;

			if (len < linear_len)
				memcpy(&buffer[tail], data, len);
				// swiCopy(data, &buffer[tail], len);
			else
			{
				memcpy(&buffer[tail], data, linear_len);
				memcpy(buffer, &data[linear_len], len-linear_len);
				// swiCopy(data, &buffer[tail], linear_len);
				// swiCopy(&data[linear_len], buffer, len-linear_len);
			}

			tail = (tail+len) % size;

			return 1;
		}

		return 0;
	}
};

int ptr_background;
int ptr_subbackground;
FILE *file_video = nullptr;
FILE *file_music = nullptr;
int file_music_len;
int file_video_len;

int framecounter = 0;

RingBuffer music_rb(MUSIC_BUFFER_SIZE);
void *music_btemp;

RingBuffer video_rb(FRAME_SIZE*CHUNK_SIZE*QUEUE_SIZE);
void *video_btemp;
void *video_decompress;
void *video_compress;

u32 timer_video_start = 0;
u32 timer_video_end = 0;
u32 timer_video_avg = 0;

int main(void) {

	videoSetMode(MODE_5_3D);
	videoSetModeSub(MODE_5_2D);

	vramSetBankA(VRAM_A_MAIN_BG);
	vramSetBankC(VRAM_C_SUB_BG);
	vramSetBankF(VRAM_F_TEX_PALETTE);

	consoleDemoInit();
	consoleDebugInit(DebugDevice_NOCASH);

	video_decompress = malloc(FRAME_SIZE*CHUNK_SIZE);
	video_compress = malloc(FRAME_SIZE*CHUNK_SIZE);
	video_btemp = malloc(FRAME_SIZE);
	music_btemp = malloc(MUSIC_BUFFER_SIZE);

	bool nitrofs = nitroFSInit(NULL);
	sassert(nitrofs, "error nitrofs");

	file_video = fopen("nitro:/image.bin", "rb");
	sassert(file_video != nullptr, "failed to load image.bin");
	fseek(file_video, 0, SEEK_END);
	file_video_len = ftell(file_video);
	fseek(file_video, 0, SEEK_SET);

	file_music = fopen("nitro:/music.raw", "rb");
	sassert(file_music != nullptr, "failed to load music.raw");
	fseek(file_music, 0, SEEK_END);
	file_music_len = ftell(file_music);
	fseek(file_music, 0, SEEK_SET);

	int pal_len;
	unsigned char* pal = Raylib::LoadFile("nitro:/palette.bin", pal_len);

	ptr_background = bgInit(3, BgType_Bmp8, BgSize_B8_256x256, 0, 0);
	ptr_subbackground = bgInitSub(3, BgType_Bmp8, BgSize_B8_256x256, VRAM_C_SUB_BG, 0);

	DC_FlushRange(pal, pal_len);
	dmaCopy(pal, BG_PALETTE, pal_len);
	dmaCopy(pal, BG_PALETTE_SUB, pal_len);

	free(pal);

	Raylib::nocashMessageFormat("preload video");
	while(video_rb.size-video_rb.count >= FRAME_SIZE*CHUNK_SIZE) {
		LoadNextChunk();
		video_rb.write((char*)video_decompress, FRAME_SIZE*CHUNK_SIZE);
	}

	{
		Raylib::nocashMessageFormat("preload music");
		int len = music_rb.size - music_rb.count;
		fread(music_btemp, sizeof(char), len, file_music);
		music_rb.write((char*)music_btemp, len);
	}

	mm_ds_system sys;
	sys.mod_count 			= 0;
	sys.samp_count			= 0;
	sys.mem_bank			  = 0;
	sys.fifo_channel		= FIFO_MAXMOD;
	mmInit( &sys );

	mm_stream mystream;
	mystream.sampling_rate	= 22050;
	mystream.buffer_length	= MMSTREAM_BUF_SIZE;
	mystream.callback		= on_stream_request;
	mystream.format			= MM_STREAM_16BIT_MONO;
	mystream.timer			= MM_TIMER2;
	mystream.manual			= false;

	timerStart(3, ClockDivider_1024, TIMER_FREQ_1024(FRAMERATE), TimerCallback);
	cothread_create(ThreadEntrypoint, NULL, 0, COTHREAD_DETACHED);
	cothread_create(ThreadVideo, NULL, 0, COTHREAD_DETACHED);

	mmStreamOpen( &mystream );

	cpuStartTiming(0);

	while(1) {
		cothread_yield_irq(IRQ_VBLANK);
		scanKeys();

		if (keysUp() & KEY_A) break;

		printf("\033[4;2H\033[Kvideo buffer: %i", video_rb.count);
    printf("\033[5;2H\033[Kmusic buffer: %i", music_rb.count);
    printf("\033[6;2H\033[Kprogress: %0.f%%", ((float)framecounter/VIDEO_TOTALFRAME)*100);

		static int frame = 0;
		frame++;

		if (frame % 30 == 0)
    {
      printf("\033[8;2H\033[Kavg decoding time: %" PRIu32 "ms", timerTicks2msec(timer_video_avg/30));
      timer_video_avg = 0;
    }
	}

	return 0;
}

int ThreadVideo(void *arg)
{
	while(1) {
		cothread_yield();

		if (file_video != nullptr && video_rb.size-video_rb.count >= FRAME_SIZE*CHUNK_SIZE)
		{
			timer_video_start = cpuGetTiming();
			
			LoadNextChunk();
			video_rb.write((char*)video_decompress, FRAME_SIZE*CHUNK_SIZE);
			
			timer_video_end = cpuGetTiming();
			timer_video_avg += timer_video_end - timer_video_start;

			if (ftell(file_video) == file_video_len)
			{
				fclose(file_video);
				file_video = nullptr;
			}
		}
	}

	return 0;
}

int ThreadEntrypoint(void *arg)
{
	while(1) {
		cothread_yield();

		if (file_music != nullptr && music_rb.count < MUSIC_BUFFER_SIZE/2)
		{
			int len = music_rb.size - music_rb.count;

			if ((file_music_len - ftell(file_music)) < len)
			{
				memset(music_btemp, 0, MUSIC_BUFFER_SIZE);
				fread(music_btemp, sizeof(char), (file_music_len - ftell(file_music)), file_music);
				fclose(file_music);
				file_music = nullptr;
			} else
				fread(music_btemp, sizeof(char), len, file_music);
			
			music_rb.write((char*)music_btemp, len);
		}
	}
	
	return 0;
}

void FrameStep()
{
	if (video_rb.count)
	{
		video_rb.read((char*)video_btemp, FRAME_SIZE);
		dmaCopyAsynch(video_btemp, bgGetGfxPtr(ptr_background), FRAME_SIZE);
		framecounter++;
	}
}

void TimerCallback()
{
	FrameStep();
}

void LoadNextChunk()
{
	uint32_t compress_size;
	fread(&compress_size, sizeof(uint32_t), 1, file_video);

	fread(video_compress, sizeof(uint8_t), compress_size, file_video);
	fastlz_decompress(video_compress, compress_size, video_decompress, FRAME_SIZE*CHUNK_SIZE);
}

mm_word on_stream_request( mm_word length, mm_addr dest, mm_stream_formats format ) {
	size_t samplesize = 1;
	switch(format){
		case MM_STREAM_8BIT_MONO: samplesize = 1; break;
		case MM_STREAM_8BIT_STEREO: samplesize = 2; break;
		case MM_STREAM_16BIT_MONO: samplesize = 2; break;
		case MM_STREAM_16BIT_STEREO: samplesize = 4; break;
	}
	int len = length*samplesize;

	if (music_rb.count >= len)
	{
		music_rb.read((char*)dest, len);
	}

	if (music_rb.count < len && file_music == nullptr) mmStreamClose();

	return length;
}